from configparser import ConfigParser
import logging
from pprint import pformat

import tator

from hms_import.util import validate_type_from_config
from hms_import.o2.upload_videos import upload_videos
from hms_import.o2.upload_metadata import upload_metadata
from hms_import.summary_image import create_summary_image


logger = logging.getLogger(__name__)


def main(config_file: str):
    # Read and parse the config file
    logger.info("Parsing config file %s", config_file)
    config = ConfigParser()
    config.read(config_file)

    # Validate config values
    host = config["Tator"]["Host"]
    token = config["Tator"]["Token"]
    project_id = config["Tator"].getint("ProjectId", -1)

    tator_api = tator.get_api(host=host, token=token)
    if project_id < 0:
        raise ValueError(f"Missing ProjectId value in the Tator section of the config file")
    try:
        tator_api.get_project(project_id)
    except Exception as exc:
        raise ValueError(f"Could not get project {project_id}") from exc

    skip_decrypt = config["Tator"].getboolean("SkipDecrypt", False)
    file_type = validate_type_from_config(tator_api, config, "FileType", project_id, "file_type")
    media_type = validate_type_from_config(
        tator_api, config, "MediaType", project_id, "media_type", list_filters=[("dtype", "video")]
    )
    summary_type = validate_type_from_config(
        tator_api,
        config,
        "SummaryType",
        project_id,
        "media_type",
        list_filters=[("dtype", "image")],
    )
    directory = config["Local"]["Directory"]

    # Create trip summary image
    create_summary_kwargs = {
        "host": host,
        "media_type": summary_type,
        "directory": directory,
        "import_type": "O2",
        "skip_decrypt": skip_decrypt,
        "hdd_sn": config["Trip"].get("HddSerialNumber", None),
    }
    logger.info("Creating trip summary with configuration %s", pformat(create_summary_kwargs))
    section, vessel_name = create_summary_image(token=token, **create_summary_kwargs)
    if section:
        logger.info("Created trip summary in section %s", section)
    else:
        raise RuntimeError("Could not create trip summary and section")

    # Construct metadata argument dictionary for logging (do not log token for security)
    file_ids = []
    if skip_decrypt:
        logger.info("Skipping metadata upload for legacy format")
    else:
        upload_metadata_kwargs = {
            "host": host,
            "project_id": project_id,
            "directory": directory,
            "meta_ext": config["Local"]["MetadataExtension"],
            "file_type": file_type,
        }
        logger.info(
            "Starting metadata upload with configuration %s", pformat(upload_metadata_kwargs)
        )
        file_ids = upload_metadata(token=token, **upload_metadata_kwargs)
        logger.info("Metadata upload complete!")

    # Construct video argument dictionary for logging (do not log token for security)
    upload_video_kwargs = {
        "host": host,
        "directory": directory,
        "project_id": project_id,
        "file_ids": file_ids,
        "media_ext": config["Local"]["MediaExtension"],
        "media_type": media_type,
        "section": section,
        "vessel_name": vessel_name,
        "skip_decrypt": skip_decrypt,
    }
    logger.info("Starting video upload with configuration %s", pformat(upload_video_kwargs))
    media_ids = upload_videos(token=token, **upload_video_kwargs)
    logger.info("Video upload complete!")
    logger.info("Created the following media: %s", pformat(media_ids))

    if skip_decrypt:
        logger.info("Skipping launch of decryption workflow!")
    else:
        logger.info(
            "Launching decryption workflow for %d media objects and %d metadata files",
            len(media_ids),
            len(file_ids),
        )

        # Launch one workflow for all media ids
        algorithm_name = config["Tator"]["AlgorithmName"]
        job_spec = {"algorithm_name": algorithm_name, "media_ids": media_ids}
        try:
            response = tator_api.create_job_list(project=project_id, job_spec=job_spec)
        except Exception:
            logger.error(
                "Could not launch job with job_spec=%s in project %d",
                pformat(job_spec),
                project_id,
                exc_info=True,
            )
        else:
            logger.info(
                "Launched workflow %s on media %s (received response %s)",
                algorithm_name,
                pformat(media_ids),
                pformat(response),
            )



from argparse import ArgumentParser, RawDescriptionHelpFormatter
from io import TextIOBase
import logging
from logging.handlers import TimedRotatingFileHandler
import os
import sys

import tator

from hms_import.util import file_exists


def get_parser():
    parser = ArgumentParser(
        description="Script for importing video and metadata in O2 and B3 formats.",
        formatter_class=RawDescriptionHelpFormatter,
    )
    parser.add_argument(
        "--quiet",
        action="store_true",
        help="Changes the console log level from INFO to WARNING; defers to --verbose",
    )
    parser.add_argument(
        "--verbose",
        action="store_true",
        help="Changes the console log level from INFO to DEBUG; takes precedence over --quiet",
    )
    cmd_parser = parser.add_subparsers(title="Commands", dest="command")

    b3_parser = cmd_parser.add_parser(
        "b3-upload", help="Imports video and GPS files from unlocked LUKS-encrypted device"
    )
    b3_parser = tator.get_parser(b3_parser)
    b3_parser.add_argument("--media-type-id", type=int)
    b3_parser.add_argument("--file-type-id", type=int)
    b3_parser.add_argument("--multi-type-id", type=int)
    b3_parser.add_argument("--state-type-id", type=int)
    b3_parser.add_argument("--image-type-id", type=int)
    b3_parser.add_argument("--directory", type=str)
    b3_parser.add_argument("--hdd-sn", type=str, required=False)

    o2_parser = cmd_parser.add_parser(
        "o2-upload", help="Script for uploading raw, encrypted video files."
    )
    o2_parser.add_argument(
        "config_file", type=str, help=f"The configuration .ini file used to initialize {__name__}."
    )

    log_parser = cmd_parser.add_parser(
        "log-upload",
        help=f"Uploads the log file {HmsLogHandler.log_filename} to Tator, if it exists",
    )
    log_parser = tator.get_parser(log_parser)
    log_parser.add_argument("--log-file-type-id", type=int)
    log_parser.add_argument(
        "--log-filename", type=file_exists, required=False, default=HmsLogHandler.log_filename
    )
    return parser


class HmsLogHandler(TextIOBase):
    log_filename = os.path.join(os.getcwd(), f"{__name__.split('.')[0]}.log")

    def __init__(self, console_log_level, *args, **kwargs):
        super().__init__(*args, **kwargs)
        root_logger = logging.getLogger()
        root_logger.setLevel(logging.DEBUG)
        self._console_handler = logging.StreamHandler()
        self._console_handler.setLevel(console_log_level)
        self._file_handler = TimedRotatingFileHandler(
            HmsLogHandler.log_filename, when="midnight", interval=1, backupCount=7
        )
        self._file_handler.setLevel(logging.DEBUG)
        formatter = logging.Formatter("[%(asctime)s - %(levelname)s - %(name)s]: %(message)s")
        self._console_handler.setFormatter(formatter)
        self._file_handler.setFormatter(formatter)
        root_logger.addHandler(self._console_handler)
        root_logger.addHandler(self._file_handler)
        self._stdout = sys.stdout
        self._stderr = sys.stderr

    def write(self, *args, **kwargs):
        self._console_handler.stream.write(*args, **kwargs)
        self._file_handler.stream.write(*args, **kwargs)

    def __enter__(self):
        sys.stdout = self
        sys.stderr = self

    def __exit__(self, *_):
        sys.stdout = self._stdout
        sys.stderr = self._stderr


def main() -> None:
    # Parse arguments
    parser = get_parser()
    args = parser.parse_args()
    argdict = vars(args)

    # Always log everything to file, set console log level based on `--quiet` and `--verbose` flags
    console_log_level = logging.INFO
    if argdict.pop("quiet"):
        console_log_level = logging.WARNING
    if argdict.pop("verbose"):
        console_log_level = logging.DEBUG

    # Import the desired main function
    with HmsLogHandler(console_log_level=console_log_level):
        command = argdict.pop("command")
        if command == "o2-upload":
            from hms_import.o2 import main
        elif command == "b3-upload":
            from hms_import.b3 import main
        elif command == "log-upload":
            from hms_import.logs import main
        else:
            raise RuntimeError(f"Received unhandled command '{command}'")

        main(**argdict)



from glob import glob
import logging
import os
from typing import List

import tator
from tator.openapi.tator_openapi.models import File
from tqdm import tqdm

logger = logging.getLogger(__name__)


def _import_metadata(host: str, token: str, project_id: int, filename: str, file_type: int) -> int:
    tator_api = tator.get_api(host=host, token=token)

    # Perform idempotency check
    base_filename = os.path.basename(filename)
    paginator = tator.util.get_paginator(tator_api, "get_file_list")
    page_iter = paginator.paginate(project=project_id)
    try:
        for page in page_iter:
            for file_obj in page:
                if file_obj.name == base_filename:
                    file_id = file_obj.id
                    logger.info("Found existing file %d with same name, skipping upload", file_id)
                    return file_id
    except RuntimeError:
        # Bug in paginator will raise RuntimeError if zero entities are returned
        pass

    # Upload encrypted file to Tator
    response = None
    file_id = 0

    try:
        for progress, response in tator.util.upload_generic_file(
            tator_api, file_type, filename, "Encrypted sensor data", name=base_filename
        ):
            logger.info("Upload progress: %.1f%%", progress)
    except Exception as exc:
        raise RuntimeError(f"Raised exception while uploading '{base_filename}', skipping") from exc

    if isinstance(response, File) and isinstance(response.id, int):
        file_id = response.id

    return file_id


def upload_metadata(
    *, host: str, token: str, project_id: int, directory: str, meta_ext: str, file_type: int
) -> List[int]:
    """Finds all encrypted metadata files in `directory`, uploads them to Tator, and runs a workflow
    to decrypt and turn them into States. Disallows use of positional arguments.

    :param host: The hostname of the Tator deployment to upload to.
    :type host: str
    :param token: The Tator API token to use for authentication.
    :type token: str
    :param project_id: The integer id of the project to upload the videos to.
    :type project_id: int
    :param directory: The directory to search for encrypted metadata files.
    :type directory: str
    :param meta_ext: The extension of the encrypted metadata files.
    :type meta_ext: str
    """
    file_list = glob(os.path.join(directory, f"*{meta_ext}-[0-9]*"))
    logger.debug("Found the following files:\n* %s", "\n* ".join(file_list))

    results = []
    for filename in tqdm(
        file_list,
        total=len(file_list),
        desc="GPS Imports",
        dynamic_ncols=True,
        position=0,
        ascii=False,
    ):
        try:
            file_id = _import_metadata(host, token, project_id, filename, file_type)
        except Exception:
            logger.error("Failed to import '%s'", os.path.basename(filename), exc_info=True)
        else:
            if file_id:
                results.append(file_id)

    return results
